import java.io.File;
import java.io.FileWriter;
import java.util.Random;
import java.util.Scanner;

public class DataWorker {
    private GraphPoint2D[] points;
    public GraphPoint3D[] holedColors;
    private final Scanner sc = new Scanner(System.in);
    private final Random rand = new Random();

    /******
     * DataWorker constructor, generates points, holes, loads data, prints help
     */
    public DataWorker(){}

    /*****
     * method loading data from a file
     * @param fileName name of file with data
     */
    public GraphPoint2D[] loadData(String fileName){
        File file = new File(fileName);
        try {
            if (file.createNewFile()) {
                System.out.println("ERROR - File \" " + fileName + " \" not found");
                file.delete();
                printHelp();
                System.exit(-2);
            }
            else {
                Scanner fileReader = new Scanner(file);
                if (!fileReader.hasNext()){
                    System.out.println("ERROR - File is empty");
                    System.exit(-2);
                }

                int numOfPoints = 0;
                String line = fileReader.nextLine();
                try {
                    numOfPoints = Integer.parseInt(line);
                }
                catch (Exception e){
                    System.out.println("ERROR - File is in wrong format");
                    System.exit(-2);
                }
                this.points = new GraphPoint2D[numOfPoints];
                for (int i = 0; i < numOfPoints; i++) {
                    if (!fileReader.hasNext()){
                        System.out.println("ERROR - File - not enough points given");
                        System.exit(-2);
                    }
                    line = fileReader.nextLine();
                    String[] parts = line.split(" ");
                    this.points[i] = new GraphPoint2D(Double.parseDouble(parts[0]), Double.parseDouble(parts[1]));
                }
            }
        }

        catch(Exception exception){
            System.out.println("ERROR - DATA LOADING, File error");
            exception.printStackTrace();
        }
        System.out.println("Loading finished");
        return this.points;
    }

    /**********
     *
     * @param input string from command line (path or num. of points to generate)
     * @return array of points - loaded or generated
     */
    public GraphPoint2D[] getPoints(String input){ ;
        boolean from_file = false;
        try {
            int num_of_points = Integer.parseInt(input);
            if (num_of_points <= 0){
                System.out.println("At least some points are needed. Please, try again.");
                printHelp();
                System.exit(-1);
            }
            generateRandomPoints(num_of_points);
        } catch (Exception exception) {
            from_file = true;
        }
        if (from_file)
            this.points = loadData(input);
        return points;
    }

    /**********
     * prints all possible orders in right order to command line
     */
    public void printHelp() {
        System.out.println("\n ---HELP---");
        System.out.println();
        System.out.println("1,  way/to/file.txt OR number_of_points");
        System.out.println("2,  i-num_of_iterations OR s-end_condition-control_iteration (end condition in (0,1))");
        System.out.println("3,  CREATE OR SEEK  --to create a hole or seek state of a position");
        System.out.println("4,  x_1             --coordinate of hole");
        System.out.println("5,  y_1             --coordinate of hole");
        System.out.println("6,  x_2             --coordinate of hole");
        System.out.println("7,  y_2             --coordinate of hole");
        System.out.println("\n NON-COMPULSORY PARAMS \n");
        System.out.println("p-num_of_iterations --iterations passed before print out");
        System.out.println("r-num_of_iterations --number of iterations before restart");


    }


    /**********
     * generates random points
     * @return array of points
     */
    public void generateRandomPoints(int pointNum){
        this.points = new GraphPoint2D[pointNum];

        for (int i = 0; i < pointNum; i++) {
            double x = rand.nextDouble();
            double y = rand.nextDouble();
            this.points[i] = new GraphPoint2D(x, y);

        }
        String name = pointNum + "Points";
            save_points(name, points, 1);
            //System.out.println("Points successfully saved to file");
    }

    /**********
     * saves points to a file
     * @param fileName name of file
     * @param points array of points to save
     * @param x num of times tried to save
     */
    public void save_points(String fileName, GraphPoint2D[] points, int x) {
        File file = new File(fileName + "(" + x + ").txt");
        try {
            if (file.createNewFile()) {
                System.out.println("Points saved: " + file.getName());
            } else {
                save_points(fileName, points, x+1);
                return;
            }
            FileWriter writer = new FileWriter(fileName + ".txt");
            writer.write(points.length + "\n");
            for (int i = 0; i < points.length; i++) {
                writer.write(points[i].getX() + " " + points[i].getY() + "\n");
            }
            writer.close();
        }
        catch(Exception exception){
            System.out.println("ERROR - POINTS SAVING");
            System.exit(-3);
        }
    }

    /*********
     * creates rectangle hole in data
     * @param points array of all points
     * @param topPosition left top corner of hole inside [0,1]^2 space
     * @param botPosition right bot corner of hole inside [0,1]^2 space
     * @param colors original array of colors from where are deleted ones connected to deleted points
     * @return array of data with hole
     */
    public GraphPoint2D[] createHoleInData(GraphPoint2D[] points, GraphPoint2D topPosition, GraphPoint2D botPosition, GraphPoint3D[] colors){
        int counter = 0;
        for (int i = 0; i < points.length; i++) {
            if ((points[i].getX() >= topPosition.getX() && points[i].getX() <= botPosition.getX()) && (points[i].getY() >= topPosition.getY() && points[i].getY() <= botPosition.getY())) {
                points[i].setX(-1);
                points[i].setY(-1);
                colors[i].setX(-1);
                colors[i].setY(-1);
                colors[i].setZ(-1);
                counter++;
            }
        }
        GraphPoint2D[] holedData = new GraphPoint2D[points.length - counter];
        this.holedColors = new GraphPoint3D[points.length - counter];
        counter = 0;
        for (int i = 0; i < holedData.length; i++) {
            if (points[counter].getX() == -1){
                counter++;
                i--;

            }
            else {
                holedData[i] = new GraphPoint2D(points[counter].getX(), points[counter].getY());
                holedColors[i] = new GraphPoint3D(colors[counter].getX(), colors[counter].getY(), colors[counter].getZ());
                counter++;
            }
        }
        return holedData;
    }


    /**********
     * transfers coordinate from string to double
     * @param arg string with coordinate
     * @return one coordinate
     */
    public double getcoordinate(String arg) {
        double coordinate = 0;
        try {
             coordinate = Double.parseDouble(arg);
            if (coordinate > 1 || coordinate < 0){
                System.out.println("Hole out of space");
                System.exit(-1);
            }
        }
        catch (Exception exception) {
            System.out.println("ERROR - coordinates error");
            printHelp();
            System.exit(-5);
        }
        return coordinate;
    }

    /******
     * gets from sting end condition, if possible
     * @param s string with condition
     * @return [end_condition, iteration, where condition is controlled]
     */
    public double[] getEndCondition(String[] s) {
        double condition = -1;
        double stopIterating = -1;
        double[] out;
        try {
            condition = Double.parseDouble(s[1]);
            if (condition > 1 || condition < 0){
                System.out.println("endcondition is < 0, cannot be reached");
                System.out.println("end condition needs to be between 0 and 1");
                System.exit(-1);
            }
            if (s.length == 3){
                stopIterating = Double.parseDouble(s[2]);
                out = new double[]{condition, stopIterating};
                return out;
            }

        }
        catch (Exception exception) {
            System.out.println("ERROR - endcondition error");
            System.exit(-5);
        }
        out = new double[]{condition};
        return out;
    }

    /**********
     * gets number of iterations to be done
     * @param arg string with num of iterations
     * @return number of iterations
     */
    public int getNumOfIterations(String arg) {
        int iterations = -1;
        try {
            iterations = Integer.parseInt(arg);

            }
        catch (Exception exception) {
            System.out.println("ERROR - not valid number of iterations");
            printHelp();
            System.exit(-3);
        }
        if (iterations < 0){
            System.out.println("There must be at least '0' iterations done");
            printHelp();
            System.exit(-3);
        }
        return iterations;
    }

    /**********
     * finds out, whether given order is print or reset or false one
     * @param arg string with order
     * @return 0 - print; 1 - reset; 2 - ERROR
     */
    public int printOrReset(String arg) {
        String[] test = arg.split("-");
        if (test[0].equals("p"))
            return 0;
        else if (test[0].equals("r"))
            return 1;
        else {
            System.out.println("Wrong 8th parameter, skipping.");
            return 2;
        }
    }

    /**********
     * gets on which iteration print of Voronoi diagram is needed
     * @param arg string with iteration
     * @return Integer with iteration
     */
    public int printTime(String arg) {
        int iteration = -1;
        String[] outParting = arg.split("-");
        try {
            iteration = Integer.parseInt(outParting[1]);
            if (iteration <= 0){
                System.out.println("ERROR - not valid iteration number");
                System.exit(-4);
            }
        }
        catch (Exception e){
            System.out.println("ERROR - not valid 8th parameter");
            System.exit(-4);
        }
        return iteration;
    }

    /**********
     * gets when McQueen algorithm is supposed to restart
     * @param arg string with time when to restart
     * @return iteration of restart
     */
    public int resetTime(String arg) {
        int iteration = -1;
        String[] outParting = arg.split("-");
        try {
            iteration = Integer.parseInt(outParting[1]);
            if (iteration <= 0){
                System.out.println("ERROR - not valid restart iteration");
                System.exit(-4);
            }
        }
        catch (Exception e){
            System.out.println("ERROR - not valid restart iteration");
            System.exit(-4);
        }
        return iteration;
    }
}
