import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferInt;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;

public class McQueen {
    private final int width;
    private final int height;
    private final Random rand = new Random();
    private final Random forColorsRand = new Random(12);
    private final GraphPoint2D[] points;
    public  final GraphPoint2D[] savePoints;
    private final GraphPoint3D[] colors;
    private BufferedImage bufferedImage;
    private int[][] pixels;
    private double closest_outhole_distance;
    private double longest_outhole_distance = 0;
    private double average = 0;
    private double success_rate = -1;
    private List<Integer> inHolePoints = new ArrayList<>();
    private double closest_inhole_distance;
    private double longest_inhole_distance = 0;
    private double average_inhole_distance = 0;
    private File nums;
    private FileWriter outWriter;
    private BufferedWriter bufferedWriter;
    private double x1, x2, y1, y2;

    /*********
     * constructor of McQueen class
     * @param width width of future pictures of voronoi diagram
     * @param height height of future pictures of Voronoi diagram
     * @param points array of points to work with
     * @param x1 x-axis coordinate of top left corner of hole
     * @param y1 y-axis coordinate of top left corner of hole
     * @param x2 x-axis coordinate of bottom right corner of hole
     * @param y2 y-axis coordinate of bottom right corner of hole
     */
    public McQueen(int width, int height, GraphPoint2D[] points, double x1, double y1, double x2, double y2) {
        this.width = width;
        this.height = height;
        this.pixels = new int[width][height];
        this.savePoints = points;
        this.points = new GraphPoint2D[savePoints.length];
        for (int i = 0; i < this.points.length; i++) {
            this.points[i] = new GraphPoint2D(savePoints[i].getX()*this.width, savePoints[i].getY() * this.height);
        }
        colors = new GraphPoint3D[this.points.length];
        bufferedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
        nums = new File("Statistics.txt");


        for (int i = 0; i < this.points.length; ++i) {
            colors[i] = new GraphPoint3D(this.forColorsRand.nextInt(255), this.forColorsRand.nextInt(255),
                    this.forColorsRand.nextInt(255));
        }
        this.x1 = x1;
        this.x2 = x2;
        this.y1 = y1;
        this.y2 = y2;
    }



    /************
     * constructor of McQueen class
     * @param width width of future pictures of voronoi diagram
     * @param height height of future pictures of Voronoi diagram
     * @param points array of points to work with
     */
    public McQueen(int width, int height, GraphPoint2D[] points) {
        this.width = width;
        this.height = height;
        this.pixels = new int[width][height];
        this.savePoints = points;
        this.points = new GraphPoint2D[savePoints.length];
        for (int i = 0; i < this.points.length; i++) {
            this.points[i] = new GraphPoint2D(savePoints[i].getX()*this.width, savePoints[i].getY() * this.height);
        }
        colors = new GraphPoint3D[this.points.length];
        bufferedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_RGB);
        nums = new File("Statistics.txt");

        for (int i = 0; i < this.points.length; ++i) {
            colors[i] = new GraphPoint3D(this.forColorsRand.nextInt(255), this.forColorsRand.nextInt(255),
                    this.forColorsRand.nextInt(255));
        }
    }

    /***********
     * method generating Voronoi
     */
    private void generateVoronoi() {
        int last_change_point = -1;
        int index;
        for (GraphPoint2D point: this.savePoints) {
            point.nullNeighbours();
        }

        int[] imagePixelData = ((DataBufferInt) bufferedImage.getRaster().getDataBuffer()).getData();
        Arrays.fill(imagePixelData, 0);

        for (int y = 0; y < height; y++) {
            last_change_point = -1;
            for (int x = 0; x < width; x++) {

                index = findClosestPoint(new GraphPoint2D(x, y));
                /*******/
                if (last_change_point == -1){
                    last_change_point = index;
                }
                if (index != last_change_point){
                    this.savePoints[index].addNeighbour(last_change_point);
                    this.savePoints[last_change_point].addNeighbour(index);
                    //zaznamenam si aktualni point, ktery je platny
                    last_change_point = index;
                }
                /*******/
                this.pixels[x][y] = index;
                imagePixelData[y * width + x] = (colors[index].getX() << 16) |
                        (colors[index].getY() << 8) | (colors[index].getZ());

            }
        }
        drawPoints();
    }

    /*********
     * method to draw woronoi colors to corresponding pixels
     */
    private void drawPoints() {
        for (int i = 0; i < this.points.length; ++i) {
            int pointSize = 1;
            for (int y = -pointSize; y <= pointSize; ++y) {
                for (int x = -pointSize; x <= pointSize; ++x) {
                    if (x * x + y * y <= pointSize * pointSize && x + points[i].getX() >= 0 &&
                            x + points[i].getX() < width && y + points[i].getY() >= 0 && y + points[i].getY() < height) {
                        bufferedImage.setRGB(x + (int)(points[i].getX()), y + (int)(this.points[i].getY()), 0);
                    }
                }
            }
        }
    }

    /*****
     * method finding closest point to given point in array of all points
     * @param point point to find nearest neighbour for
     * @return inder of nearest neighbour in array
     */
    private int findClosestPoint(GraphPoint2D point) {
        int index = 0;
        double minDistance = distance(point, points[index]);
        double currentDistance;
        for (int i = 1; i < this.points.length; ++i) {
            currentDistance = distance(point, points[i]);
            if (currentDistance < minDistance) {
                minDistance = currentDistance;
                index = i;
            }
        }
        return index;
    }

    /****
     * method measuring distance between two points
     * @param pointA pointA
     * @param pointB pointB
     * @return distance between pointA and pointB
     */
    private double distance(GraphPoint2D pointA, GraphPoint2D pointB) {
        return Math.sqrt((pointA.getX() - pointB.getX()) * (pointA.getX() - pointB.getX())
                + (pointA.getY() - pointB.getY()) * (pointA.getY() - pointB.getY()));
    }

    /******
     * method measuring closest and longest distance, inhole and arround
     */
    private void measure_distance(){
        this.closest_inhole_distance = Double.POSITIVE_INFINITY;
        this.longest_inhole_distance = 0;
        this.average_inhole_distance = 0;
        double distance;
        double point_neigh_dist;
        this.longest_outhole_distance = 0;
        this.closest_outhole_distance = Double.POSITIVE_INFINITY;
        this.average = 0;
        this.inHolePoints.clear();
        boolean isInHole;
        for (int i = 0; i < this.savePoints.length; i++) {
            isInHole = false;
            distance = 0;
            double x = this.savePoints[i].getX();
            double y = this.savePoints[i].getY();
            if ((x > this.x1 && x < this.x2) && (y > this.y1 && y < this.y2)) {
                this.inHolePoints.add(i);
                isInHole = true;
            }
            for (int neighbour : this.savePoints[i].getNeighbours()){
                point_neigh_dist = distance(this.savePoints[i], this.savePoints[neighbour]);
                if (point_neigh_dist < this.closest_outhole_distance && !isInHole) {
                    this.closest_outhole_distance = point_neigh_dist;
                }
                if(point_neigh_dist > this.longest_outhole_distance){
                    this.longest_outhole_distance = point_neigh_dist;
                }
                distance += point_neigh_dist;

            }
            distance /= this.savePoints[i].getNeighbours().size();
            this.average += distance;
        }

        this.average /= this.savePoints.length;

        for (int i = 0; i < this.inHolePoints.size(); i++) {
            distance = 0;
            double x = this.points[this.inHolePoints.get(i)].getX();
            double y = this.points[this.inHolePoints.get(i)].getY();
            for (int neighbour : this.savePoints[this.inHolePoints.get(i)].getNeighbours()) {
                point_neigh_dist = distance(this.savePoints[this.inHolePoints.get(i)], this.savePoints[neighbour]);
                if (point_neigh_dist < this.closest_inhole_distance) {
                    this.closest_inhole_distance = point_neigh_dist;
                }
                if (point_neigh_dist > this.longest_inhole_distance) {
                    this.longest_inhole_distance = point_neigh_dist;

                }
                distance += point_neigh_dist;

                distance /= this.savePoints[this.inHolePoints.get(i)].getNeighbours().size();
                this.average_inhole_distance += distance;
            }
        }
        this.average_inhole_distance /= this.inHolePoints.size();
        if (this.inHolePoints.size() == 0)
            this.average_inhole_distance = Double.POSITIVE_INFINITY;
    }

    /*****
     * calculates success rate, how well filled hole is
     */
    private void calculate_success_rate() {
        this.success_rate = ((double) this.inHolePoints.size() / this.savePoints.length) / (Math.abs(this.x2 - this.x1)
                * Math.abs(this.y2 - this.y1));
    }

    /*****
     * moves one point using McQueen method
     */
    public void mcQueenCentralizationOfOnePoint(){
        GraphPoint2D helpingPoint = new GraphPoint2D(rand.nextDouble() % 1, rand.nextDouble() % 1);
        int pointIndex = 0;
        for (int i = 1; i < this.savePoints.length; i++) {
            if (distance(this.savePoints[i], helpingPoint) < distance(this.savePoints[pointIndex], helpingPoint)){
                pointIndex = i;
            }
        }
        //found closest point, now move it
        this.savePoints[pointIndex].update();
        this.savePoints[pointIndex].setX((helpingPoint.getX() + (this.savePoints[pointIndex].getX()
                * this.savePoints[pointIndex].getTimesUpdated())) / (this.savePoints[pointIndex].getTimesUpdated() + 1));

        this.savePoints[pointIndex].setY((helpingPoint.getY() + (this.savePoints[pointIndex].getY()
                * this.savePoints[pointIndex].getTimesUpdated())) / (this.savePoints[pointIndex].getTimesUpdated() + 1));

        this.points[pointIndex].setX(this.savePoints[pointIndex].getX() * this.width);
        this.points[pointIndex].setY(this.savePoints[pointIndex].getY() * this.height);

    }

    /****************
     * method starting generation of Voronoi diagram and other data
     * @param printIt true -prints picture out
     * @param generateIt true - generates numbers
     * @param text text imprinted on the picture
     * @param iterationNo number of iteration info
     */
    public void generate(boolean printIt, boolean generateIt, String text, int iterationNo) {
        if (generateIt || printIt) {
            generateVoronoi();
            measure_distance();
            if (iterationNo != 0)
                calculate_success_rate();
            try {
                if (generateIt) {
                    writeoutPoints(iterationNo);
                    startWriter();
                    this.bufferedWriter.write(this.closest_outhole_distance + " " + this.longest_outhole_distance + " " +
                            this.average + " " + this.closest_inhole_distance + " " + this.longest_inhole_distance +
                            " " + this.average_inhole_distance + " " + this.success_rate + "\n");
                    closeFile();
                }
                if (printIt) {
                    if (!generateIt){
                        startWriter();
                        writeoutPoints(iterationNo);
                        this.bufferedWriter.write(this.closest_outhole_distance + " " + this.longest_outhole_distance +
                                " " + this.average + " " + this.closest_inhole_distance + " " +
                                this.longest_inhole_distance + " " + this.average_inhole_distance + this.success_rate +
                                "\n");
                        closeFile();
                    }
                    ImageIO.write(bufferedImage, "png", new File(text + ".png"));
                }

            } catch (Exception e) {
                e.printStackTrace();
            }
        }
    }

    /****
     * method writing actual points to a file in given format
     * @param iterationNo actual iteration
     */
    public void writeoutPoints(int iterationNo) {
        try {
            FileWriter points = new FileWriter("points_iterno_" + iterationNo + ".txt");
            points.write(this.savePoints.length + "\n");
            for (GraphPoint2D point2D : this.savePoints) {
                points.write(String.valueOf(point2D) + "\n");
            }
            points.close();

        } catch (Exception exception) {
            exception.printStackTrace();
        }

    }


    public GraphPoint3D[] getColors() {
        return colors;
    }

    public double getAverage() {
        return average;
    }

    public double getAverage_inhole_distance() {
        return average_inhole_distance;
    }

    /******
     * opens Statistics file for writting
     */
    private void startWriter() {
        try {
            this.outWriter = new FileWriter("Statistics.txt", true);
            this.bufferedWriter = new BufferedWriter(outWriter);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /****
     * closes a file
     */
    public void closeFile(){
        try {
            this.bufferedWriter.close();
        }
        catch (Exception exception){
            System.out.println("ERROR - File with data could not be closed.");
        }
    }


}
